# EDG_SET_VERSION_FROM_CVS_TAG(CVS_NAME, DEFAULT_VERSION)
# -------------------------------------------------------
AC_DEFUN([EDG_SET_VERSION_FROM_CVS_TAG],
[AC_MSG_CHECKING([for version])
version=`echo $1 | cut -d':' -f2 | tr '_' '.' | tr -d 'v$ '`
VERSION=${version:-$2}
AC_MSG_RESULT([$VERSION])
AC_SUBST(VERSION)])

# EDG_SET_RPM_TOPDIR(DIRECTORY)
# -----------------------------
AC_DEFUN([EDG_SET_RPM_TOPDIR],
[AC_MSG_CHECKING([for rpm topdir])
AC_ARG_WITH([rpm-dir],
            [  --with-rpm-dir=DIR      rpm topdir in DIR [`pwd`]],
            [rpm_topdir=$withval],
            [rpm_topdir=`pwd`])
if test -d $rpm_topdir ; then
   AC_MSG_RESULT([$rpm_topdir found])
else
   AC_MSG_RESULT([$rpm_topdir not found])
fi
RPM_TOPDIR=$rpm_topdir
AC_SUBST(RPM_TOPDIR)])

# SOCKLEN_T TESTING
# -----------------
AC_DEFUN([VOMS_SOCKLEN_T],
[AC_MSG_CHECKING([for (sane) socklen_t])
cat >conftest.cpp <<HERE
#include <sys/types.h> 
#include <sys/socket.h>

int main(int argc, char *argv[]) {
  socklen_t addrlen = (socklen_t)5;

  (void)getsockname(0, NULL, &addrlen);
  return 0;
}
HERE
if ( ($CXX -g -O2 -o conftest.o conftest.cpp >/dev/null 2>&1)); then
AC_DEFINE(HAVE_SOCKLEN_T, 1, HAVE_SOCKLEN_T)
AC_MSG_RESULT(yes)
else 
AC_MSG_RESULT(no)
fi
rm -rf conftest*
])


AC_DEFUN([OPENSSL_EXT_METHOD],
[AC_MSG_CHECKING([for it member in X509V3_EXT_METHOD])
CFLAGS_SAVE="$CFLAGS"
CFLAGS="$CFLAGS -I$GLOBUS_LOCATION/include/$GLOBUS_FLAVOR_NAME"
AC_COMPILE_IFELSE(AC_LANG_PROGRAM(
[[
#include <openssl/x509v3.h>
]],
[[X509V3_EXT_METHOD it; it.it = NULL;]]), 
[AC_MSG_RESULT(yes) 
  AC_DEFINE(HAVE_X509V3_EXT_METHOD_IT, 1, HAVE_X509V3_EXT_METHOD_IT)], 
[AC_MSG_RESULT(no)])
CFLAGS="$CFLAGS_SAVE"
])

# VOMS_LONG_LONG
# --------------
AC_DEFUN([VOMS_LONG_LONG],
[AC_MSG_CHECKING([for long long])
cat >conftest.c <<HERE
int main(int argc, char *argv[])
{
	long long i;
}
HERE
if ( ($CXX -c -o conftest.o conftest.c >/dev/null 2>&1)); then
AC_DEFINE(HAVE_LONG_LONG_T, 1, HAVE_LONG_LONG_T)
AC_MSG_RESULT(yes)
else
AC_MSG_RESULT(no)
fi
#rm -rf conftest*
])

# VOMS_GLOBUS_OFF_T
# ------------
AC_DEFUN([VOMS_GLOBUS_OFF_T],
[AC_MSG_CHECKING([for GLOBUS_OFF_T])
CFLAGS_SAVE="$CFLAGS"
CFLAGS="$CFLAGS -I$GLOBUS_LOCATION/include/$GLOBUS_FLAVOR_NAME"
AC_TRY_COMPILE([#include <globus_common.h>], [GLOBUS_OFF_T goff],
	[AC_DEFINE(HAVE_GLOBUS_OFF_T, 1, HAVE_GLOBUS_OFF_T)
	 AC_MSG_RESULT(yes)], [AC_MSG_RESULT(no)])
CFLAGS="$CFLAGS_SAVE"
])

# VOMS_FINDFUNC
# -------------
AC_DEFUN([VOMS_FIND_FUNC],
[AC_MSG_CHECKING([for function name discovery])
AC_TRY_COMPILE([],[char *str = __func__], [
func="__func__"], [
AC_TRY_COMPILE([],[char *str = __PRETTY_FUNCTION__], [
func="__PRETTY_FUNCTION__"], [
AC_TRY_COMPILE([],[char *str = __FUNCTION__], [
func="__FUNCTION__"], [
func=""])])])
if test "x$func" = "x"; then
AC_MSG_RESULT(no)
func="NULL"
else
AC_MSG_RESULT([yes ($func)])
fi
AC_DEFINE_UNQUOTED(FUNC_NAME, [$func], FUNC_NAME)
])

# VOMS_STRUCT_IOVEC
# -----------------
AC_DEFUN([VOMS_STRUCT_IOVEC],
[AC_MSG_CHECKING([for struct iovec])
cat >conftest.cpp <<HERE
#include <sys/uio.h>
int main(int argc, char *argv[]) {
	struct iovec v;
}
HERE
if ( ($CXX -g -O2 -o conftest.o conftest.cpp >/dev/null 2>&1)); then
AC_DEFINE(HAVE_STRUCT_IOVEC, 1, HAVE_STRUCT_IOVEC)
AC_MSG_RESULT(yes)
else
AC_MSG_RESULT(no)
fi
rm -rf conftest*
])

AC_DEFUN([VOMS_GLOBUS_CONFIG_H],
[AC_MSG_CHECKING([for globus_config.h])
cat >conftest.cpp <<HERE
#include <globus_config.h>
HERE
if ( ($CXX $CXXFLAGS -c -o conftest.o conftest.cpp >/dev/null 2>&1)); then
AC_MSG_RESULT(yes)
else
AC_MSG_RESULT([no (creating)])
proc=`./config.guess|cut -d- -f1`
arch=`./config.guess|cut -d- -f3`
case "$proc" in
  i*86) proc=X86 ;;
  ia64) proc=IA64 ;;
  *) proc="" ;;
esac
arch=`echo $arch|tr a-z A-Z`
echo "#define BUILD_LITE 1"   >include/globus_config.h
echo "#define BUILD_DEBUG 1" >>include/globus_config.h
echo "#define TARGET_ARCH_$arch 1" >>include/globus_config.h
if test "x$proc" = "x"; then
echo;
else
echo "#define TARGET_ARCH_$proc 1" >>include/globus_config.h
fi
fi
])

AC_DEFUN([VDT_CHECKS],
[VOMS_GLOBUS_CONFIG_H
VOMS_STRUCT_IOVEC
VOMS_GLOBUS_OFF_T
])

AC_DEFUN([VOMS_CHECKS],
[VOMS_SOCKLEN_T
VOMS_FIND_FUNC
VOMS_LONG_LONG
VDT_CHECKS])

AC_DEFUN([NEW_ISSUES],
[
AC_LANG_PUSH(C++)
AC_MSG_CHECKING([for string dependency on cerr])
cat >conftest.cpp <<HERE
#include <string>
int main(int argc, char *argv[]) {
	std::string g;
  return 0;
}
HERE
if ( ($CXX -c -o conftest.o conftest.cpp >/dev/null 2>&1) ); then
  if ( ($NM conftest.o | grep cerr) ); then
    AH_BOTTOM([#ifdef __cplusplus
#include <new>
#endif])
    AC_DEFINE(__THROW_BAD_ALLOC, return NULL, __THROW_BAD_ALLOC)
    AC_MSG_RESULT([yes])
  else
    AC_MSG_RESULT([no])
  fi
else
  AC_MSG_RESULT([cannot test])
fi
rm -rf conftest*
AC_LANG_POP(C++)
])

AC_DEFUN([TEST_USE_BSD],
[
AC_LANG_PUSH(C)
AC_MSG_CHECKING([wether _BSD_SOURCE must be defined])
cat >conftest.c <<HERE
#include <strings.h>
char *f(void)
{
  return strdup("try");
}
int main(int argc, char **argv) {
  (void)f();
  return 0;
}
HERE
if ( ($CC -c -o conftest.o -Wall -ansi -pedantic-errors -Werror conftest.c >/dev/null 2>&1) ); then
AC_MSG_RESULT([no])
else
  cat >conftest.c <<HERE
  #define _BSD_SOURCE
  #include <strings.h>
  char *f(void)
  {
    return strdup("try");
  }
  int main(int argc, char **argv) {
    (void)f();
    return 0;
  }
HERE
  if ( ($CC -c -o conftest.o -Wall -ansi -pedantic-errors -Werror conftest.c >/dev/null 2>&1) ); then
  AC_MSG_RESULT([Needs something else. Let's try and hope])
  else
  AC_MSG_RESULT([yes])
  AC_DEFINE(_BSD_SOURCE, 1, [needed to get ansi functions definitions])
  fi
fi
rm -rf conftest*
AC_LANG_POP(C)
])

AC_DEFUN([TEST_USE_POSIX],
[
AC_LANG_PUSH(C)
AC_MSG_CHECKING([wether _POSIX_SOURCE must be defined])
cat >conftest.c <<HERE
#include <stdio.h>
int f(void)
{
  return fileno(stderr);
}
int main(int argc, char **argv) {
  (void)f();
  return 0;
}
HERE
if ( ($CC -c -o conftest.o -Wall -ansi -pedantic-errors -Werror conftest.c >/dev/null 2>&1) ); then
AC_MSG_RESULT([no])
else
  cat >conftest.c <<HERE
  #define _POSIX_SOURCE
  #include <strings.h>
  int f(void)
  {
    return fileno(stderr);
  }
  int main(int argc, char **argv) {
    (void)f();
    return 0;
  }
HERE
  if ( ($CC -c -o conftest.o -Wall -ansi -pedantic-errors -Werror conftest.c >/dev/null 2>&1) ); then
  AC_MSG_RESULT([Needs something else. Let's try and hope])
  else
  AC_MSG_RESULT([yes])
  AC_DEFINE(_POSIX_SOURCE, 1, [needed to get ansi functions definitions])
  fi
fi
rm -rf conftest*
AC_LANG_POP(C)
])