#!/usr/bin/perl -w

# $Revision: 1.2 $
# $State: Exp $
# $Date: 2004/06/15 09:44:39 $

=head1 NAME

gen_ca_sql.pl - Fills up the VOMS.ca table

=head1 SYNOPSIS

gen_ca_sql.pl [<certificates directory>]

=head1 DESCRIPTION

This program generates SQL commands to fill up the VOMS.ca table
initially. It shall not be run later!

It uses as default C</etc/grid-security/certificate> directory,
but it can be changed by providing a parameter.

The program searches the directory for files with '.0' ending,
extracts the subject name using OpenSSL and emits the proper
SQL commands for the database initialisation.

=head1 AUTHORS

  Copyright (c) 2002 CERN, ELTE, on behalf of the EU DataGrid.
  For license conditions see LICENSE file or http://www.edg.org/license.html

=cut

use strict;
use DirHandle;

my $default_certdir = '/etc/grid-security/certificates';
my %certificates = ();
my $SSLPROG = 'openssl';
my $verbose = 0;

##############################################################
# extracts the subject and adds it to %certificates
sub extractSubject ($) {
	my ($filename) = @_;
	print "Processing certificate: $filename\n" if $verbose;
	chomp(my $subject = `$SSLPROG x509 -noout -subject -in $filename`);
	$subject =~ s/^subject=\s+//;
	
	my ($cn) = ($subject =~ /\/CN=([^\/]+)/);
	
	print "Subject: $subject, CN: $cn\n" if $verbose;

	$certificates{$cn} = $subject;
}

##############################################################
# generates the SQL commands from %certificates
sub generateSQL {
	foreach my $cn (sort(keys(%certificates))) {
		my $subject = $certificates{$cn};
		print "INSERT INTO ca (ca, cadescr) VALUES ('$subject','$cn');\n"
	}
}

##############################################################
# scans the certificate directory for .0 files and calls extractSubject
sub scanCertDir ($) {
	my ($certdir) = @_;
	
	my $d = new DirHandle($certdir) || die "Couldn't scan $certdir";
	my $filename;
	while(defined($filename = $d->read())) {
		next unless $filename =~ /\.0$/;
		extractSubject($certdir . '/' . $filename);
	}
}


##############################################################
# main
##############################################################
if(@ARGV) {
	foreach my $certdir (@ARGV) {
		next unless -d $certdir;
		scanCertDir($certdir);
	}
}
else {
	scanCertDir($default_certdir);
}

generateSQL();

